//#include "cordic-16bits.h"
#include <stdio.h>
#include <string.h>
#include <math.h> // for testing only!
#define M_PI 3.1415926536897932384626
#define K1 0.6072529350088812561694

int cordic_ctab [] = {0x00001680, 0x00000D48, 0x00000705, 0x00000390, 0x000001CA, 0x000000E5, 0x00000073, 0x00000039, 0x0000001D, 0x0000000E, 0x00000007, 0x00000004, 0x00000002, 0x00000001, 0x00000000, 0x00000000};

typedef struct record {
	short x,y,z;
	short theta, whi, xy_length, length;
	double m_theta, m_whi, d_theta, d_whi;
} RECORD;

void test_cordic_vec(short x_in, short y_in, short *l, short *a, int n)
{
  int k;
  short tx, ty, tz, d;
  short x, y, z;
  int iter_no;
  int freezed;
  double m_theta;

//  iter_no = (n>CORDIC_NTAB) ? CORDIC_NTAB : n;
  x = x_in;
  y = y_in; 
  z = 0;
  tz = 0;
  freezed = 0;
  iter_no = n;

  printf("Starting cordic for [%d,%d] with %d iterations\n", x, y, n);
  // Let's play smart, if y = 0, don't calculate
  if (x == 0) {
	  if (y > 0) {
		  *l = y, *a = 11520;
	  } else if (y < 0) {
		  *l = -y, *a = -11520;
	  } else {
		  *l = 0, *a = 0;
	  }
  } else {
	for (k=0; k<iter_no; ++k)
	{
		//if(y == 0) freezed = 1;
		d = y < 0 ? 1 : -1;
		if(freezed != 1) {
		if(d == 1) {
			tx = x - (y >> k);
			ty = y + (x >> k);
			if(freezed == 0) tz = z - cordic_ctab[k];
		} else {
			tx = x + (y >> k);
			ty = y - (x >> k);
			if(freezed == 0) tz = z + cordic_ctab[k];
		}
		}
		//if(((ty == y) && (tx == x)) || (y == -1)) freezed = 1;
		//else { 
		z = tz;
		x = tx; y = ty;
		//}
		printf("#%d=>d=%d, x=%d, y=%d, z=%d\n", k, d, x, y, z);
	 }  
	*l = (short)((float)x * K1); *a = z;
  }
  m_theta = atan((double)y_in / (double)x_in)/M_PI*180.0;
  printf("End of Codic with length=%f, angle=%f, theta=%f\n", *l/16.0, *a/128.0, m_theta);
 
}

void cordic_vec(short x_in, short y_in, short *l, short *a, int n)
{
  int k;
  short tx, ty, tz, d;
  short x, y, z;
  int iter_no;
  int freezed;

//  iter_no = (n>CORDIC_NTAB) ? CORDIC_NTAB : n;
  x = x_in;
  y = y_in; 
  z = 0;
  tz = 0;
  freezed = 0;
  iter_no = n;

//  printf("Starting cordic for [%d,%d] with %d iterations\n", x, y, n);
  // Let's play smart, if y = 0, don't calculate
  if (x == 0) {
	  if (y > 0) {
		  *l = y, *a = 11520;
	  } else if (y < 0) {
		  *l = -y, *a = -11520;
	  } else {
		  *l = 0, *a = 0;
	  }
  } else {
	for (k=0; k<iter_no; ++k)
	{
		if(y == 0) freezed = 1;
		d = y < 0 ? 1 : -1;
		if(d == 1) {
			tx = x - (y >> k);
			ty = y + (x >> k);
			if(freezed == 0) tz = z - cordic_ctab[k];
		} else {
			tx = x + (y >> k);
			ty = y - (x >> k);
			if(freezed == 0) tz = z + cordic_ctab[k];
		}
//		if((ty == y) || (ty == 0)) freezed = 1;
//		else z = tz;
		x = tx; y = ty; z = tz;
//		printf("#%d=>d=%d, x=%d, y=%d, z=%d\n", k, d, x, y, z);
	 }  
	*l = (short)((float)x * K1); *a = z;
  }
//  printf("End of Codic with length=%d, angle=%d\n", *l, *a);
}

int main(int argc, char **argv)
{
    int iters,i;    
	short x, y, z, x_max, y_max, theta, whi, length, length1;
	double m_theta, m_whi, d_theta, d_whi;
	double std_theta_error = 0.4476;
	double std_whi_error = 0.4476;
	long int no_of_tests, no_of_theta_exceed, no_of_whi_exceed;

	struct record d_whi_max, d_whi_min, d_theta_max, d_theta_min;

	if(argc == 1) iters = 8;
	else iters = atoi(argv[1]);
	
	std_theta_error = atan(pow(2, -iters))/M_PI * 180 * 2.0;
	std_whi_error = std_theta_error;

	d_whi_max.d_whi = 0.0;
	d_whi_min.d_whi = 0.0;
	d_theta_max.d_theta = 0.0;
	d_theta_min.d_theta = 0.0;
	
	test_cordic_vec(6 << 4, 4 << 4, &length, &theta, 10);
//	printf("Length = %f, and angle = %f\n", (float)length/16.0, (float)theta/128.0);

#if 0
	x_max = 255;
	no_of_tests = 0;
	no_of_theta_exceed = 0;
	no_of_whi_exceed = 0;
//	printf("x:\ty:\tz:\txy_leng:\tm_xy_leng:\tlength:\tm_length:\ttheta:\tm_theta:\td_theta:\twhi:\tm_whi:\td_whi:\n");
	for(x = 0; x <= x_max; x++) {
//	for(x = 0; x <= 10; x++) {
//		printf("\n");
		y_max = (short) sqrt(255*255 - (double)x * (double)x);
		for(y = -y_max; y <= y_max; y++) {
//		for(y = -10; y <= 10; y++) {
			z = (short) sqrt(255*255 - (double)x * (double)x - (double)y * (double)y);
//			printf("starting angle calculations for [%d, %d, %d]\n", x, y, z);
			if(x == 0) {
				if(y > 0) m_theta = 90.0;
				else if(y == 0) m_theta = 0;
				else m_theta = -90;
			} else {
				m_theta = atan((double)y / (double)x)/M_PI*180.0;
			}
			if(sqrt((double)x * (double)x + (double)y * (double)y) == 0) {
				if(z > 0) m_whi = 90.0;
				else if (z == 0) m_whi = 0;
				else m_whi = -90.0;
			} else {
				m_whi = atan((double)z / sqrt((double)x * (double)x + (double)y * (double)y))/M_PI*180.0;
			}
			
			cordic_vec(x << 4, y << 4, &length, &theta, iters);
			cordic_vec(length, z << 4, &length1, &whi, iters);
			d_theta = (float)theta/128.0 - m_theta;
			d_whi = (float)whi/128.0 - m_whi;
			
			no_of_tests++;
			if(fabs(d_theta) > std_theta_error) no_of_theta_exceed++;
			if(fabs(d_whi) > std_whi_error) no_of_whi_exceed++;

			if(d_whi > d_whi_max.d_whi) {
				d_whi_max.x = x;
				d_whi_max.y = y;
				d_whi_max.z = z;
				d_whi_max.xy_length = length;
				d_whi_max.length = length1;
				d_whi_max.theta = theta;
				d_whi_max.whi = whi;
				d_whi_max.m_theta = m_theta;
				d_whi_max.d_theta = d_theta;
				d_whi_max.m_whi = m_whi;
				d_whi_max.d_whi = d_whi;
			}
			
			if(d_theta > d_theta_max.d_theta) {
				d_theta_max.x = x;
				d_theta_max.y = y;
				d_theta_max.z = z;
				d_theta_max.xy_length = length;
				d_theta_max.length = length1;
				d_theta_max.theta = theta;
				d_theta_max.whi = whi;
				d_theta_max.m_theta = m_theta;
				d_theta_max.d_theta = d_theta;
				d_theta_max.m_whi = m_whi;
				d_theta_max.d_whi = d_whi;
			}
			
			if(d_whi < d_whi_min.d_whi) {
				d_whi_min.x = x;
				d_whi_min.y = y;
				d_whi_min.z = z;
				d_whi_min.xy_length = length;
				d_whi_min.length = length1;
				d_whi_min.theta = theta;
				d_whi_min.whi = whi;
				d_whi_min.m_theta = m_theta;
				d_whi_min.d_theta = d_theta;
				d_whi_min.m_whi = m_whi;
				d_whi_min.d_whi = d_whi;
			}
			
			if(d_theta < d_theta_min.d_theta) {
				d_theta_min.x = x;
				d_theta_min.y = y;
				d_theta_min.z = z;
				d_theta_min.xy_length = length;
				d_theta_min.length = length1;
				d_theta_min.theta = theta;
				d_theta_min.whi = whi;
				d_theta_min.m_theta = m_theta;
				d_theta_min.d_theta = d_theta;
				d_theta_min.m_whi = m_whi;
				d_theta_min.d_whi = d_whi;
			}
//			printf("%e\t", fabs(d_theta));

			printf("%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%e\t%f\t%f\t%e\n",x, y, z, (float)length/16.0, sqrt((double)x * (double)x + (double)y * (double)y), 
				(float)length1/16.0, sqrt((double)x * (double)x + (double)y * (double)y + (double)z * (double)z), 
				(float)theta/128.0, m_theta, d_theta, (float)whi/128.0, m_whi, d_whi);
			
			z = -z;
			if(sqrt((double)x * (double)x + (double)y * (double)y) == 0) {
				if(z > 0) m_whi = 90.0;
				else if (z == 0) m_whi = 0;
				else m_whi = -90.0;
			} else {
				m_whi = atan((double)z / sqrt((double)x * (double)x + (double)y * (double)y))/M_PI*180.0;
			}

			cordic_vec(length, z << 4, &length1, &whi, iters);
			d_whi = (float)whi/128.0 - m_whi;

			no_of_tests++;
			if(fabs(d_theta) > std_theta_error) no_of_theta_exceed++;
			if(fabs(d_whi) > std_whi_error) no_of_whi_exceed++;

			if(d_whi > d_whi_max.d_whi) {
				d_whi_max.x = x;
				d_whi_max.y = y;
				d_whi_max.z = z;
				d_whi_max.xy_length = length;
				d_whi_max.length = length1;
				d_whi_max.theta = theta;
				d_whi_max.whi = whi;
				d_whi_max.m_theta = m_theta;
				d_whi_max.d_theta = d_theta;
				d_whi_max.m_whi = m_whi;
				d_whi_max.d_whi = d_whi;
			}
			if(d_theta > d_theta_max.d_theta) {
				d_theta_max.x = x;
				d_theta_max.y = y;
				d_theta_max.z = z;
				d_theta_max.xy_length = length;
				d_theta_max.length = length1;
				d_theta_max.theta = theta;
				d_theta_max.whi = whi;
				d_theta_max.m_theta = m_theta;
				d_theta_max.d_theta = d_theta;
				d_theta_max.m_whi = m_whi;
				d_theta_max.d_whi = d_whi;
			}
			if(d_whi < d_whi_min.d_whi) {
				d_whi_min.x = x;
				d_whi_min.y = y;
				d_whi_min.z = z;
				d_whi_min.xy_length = length;
				d_whi_min.length = length1;
				d_whi_min.theta = theta;
				d_whi_min.whi = whi;
				d_whi_min.m_theta = m_theta;
				d_whi_min.d_theta = d_theta;
				d_whi_min.m_whi = m_whi;
				d_whi_min.d_whi = d_whi;
			}
			if(d_theta < d_theta_min.d_theta) {
				d_theta_min.x = x;
				d_theta_min.y = y;
				d_theta_min.z = z;
				d_theta_min.xy_length = length;
				d_theta_min.length = length1;
				d_theta_min.theta = theta;
				d_theta_min.whi = whi;
				d_theta_min.m_theta = m_theta;
				d_theta_min.d_theta = d_theta;
				d_theta_min.m_whi = m_whi;
				d_theta_min.d_whi = d_whi;
			}
//			printf("[%d, %d, %d] => Length=%f (%f), theta=%f (%f)\n", x, y, z, (float)length/16.0, sqrt((double)x * (double)x + (double)y * (double)y), (float)theta/128.0, m_theta);
//			printf("\t\t => Length1=%f (%f), whi=%f (%f)\n", (float)((short)((float)length1 * K1))/16.0, sqrt((double)x * (double)x + (double)y * (double)y + (double)z * (double)z), (float)whi/128.0, m_whi);
	
			printf("%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%e\t%f\t%f\t%e\n",x, y, z, (float)length/16.0, sqrt((double)x * (double)x + (double)y * (double)y), 
				(float)length1/16.0, sqrt((double)x * (double)x + (double)y * (double)y + (double)z * (double)z), 
				(float)theta/128.0, m_theta, d_theta, (float)whi/128.0, m_whi, d_whi);
	
//			getchar();

		}       
	}
//	for(i=0; i<16; i++) fprintf(stderr, "i=%d, theta=%f\n", i, atan(pow(2, -i))/M_PI * 180);
	fprintf(stderr, "%d, %d, %d, %f, %f\n", d_theta_max.x, d_theta_max.y, d_theta_max.z, d_theta_max.d_theta, d_theta_max.d_whi);
	fprintf(stderr, "%d, %d, %d, %f, %f\n", d_theta_min.x, d_theta_min.y, d_theta_min.z, d_theta_min.d_theta, d_theta_min.d_whi);
	fprintf(stderr, "%d out of %d theta = %f exceed standard error = %f\n", no_of_theta_exceed, no_of_tests, (float)no_of_theta_exceed / (float)no_of_tests, std_theta_error);
	fprintf(stderr, "%d, %d, %d, %f, %f\n", d_whi_max.x, d_whi_max.y, d_whi_max.z, d_whi_max.d_theta, d_whi_max.d_whi);
	fprintf(stderr, "%d, %d, %d, %f, %f\n", d_whi_min.x, d_whi_min.y, d_whi_min.z, d_whi_min.d_theta, d_whi_min.d_whi);
	fprintf(stderr, "%d out of %d whi = %f exceed standard error = %f\n", no_of_whi_exceed, no_of_tests, (float)no_of_whi_exceed / (float)no_of_tests, std_whi_error);
#endif
}
